﻿/**************************************************************************
 * BLH.cs
 * 緯度経度表現の座標のための構造体 for C#
 * 
 * 開発者　：森下功啓（K.Morhista Kumamoto-University）
 * 開発履歴：2011/7/3   整備開始
 *           2011/7/6   2地点間の距離を求める、GetDistanceのデバッグ。単位長までしか計算していなかった…。
 *           2011/7/17  XYZとBLHに二項演算子を追加
 *           2011/9/19  Open(string fname)のバグを修正
 *                      GetPositions(),GetPositioningResults()において、ubloxのNMEAフォーマットに対応（現時点での要求仕様を満たすだけなので万能ではないことに注意）
 *                      PositioningInfosのToString()に高度も含まれるように変更
 *           2011/11/11 よく分かっていないが、シリアル化した方が良さそうとのことで[Serializable]を片っ端からつけてみた。今後どうなるかは・・・よく分からない。
 *           2011/11/20 BLH構造体にToString()を追加
 *           2011/12/2  XML構文エラーがある部分を訂正した
 *           2011/12/15 BLH構造体にToString()を追加
 *           2012/4/22  コメントを少しだけ付け直した。
 *                      NMEAクラスは現時点ではバイナリの処理が不可能だが、そこんとこを何とかしたい。
 *           2012/4/24  ソースコードを分割した。
 *                      プロパティになっていて不自然な感じがするものはメソッドへ変更した。
 *                      測地系を変更できるような準備を導入した。が、発展させるかは不明。。。
 *           2012/4/29  Equals()を実装した。
 *           2012/5/11  ToXYZ()に、変換後の測地系を省略できるバージョンを整備した。
 *                      FilterAsLatitudeAndLongitude()を整備して、座標同士の足し算・引き算で極や経度180°をまたいだ演算を安全に行えるようにした。
 *           2012/5/14  2地点間の中間座標を返すGetMedian()に有ったバグを修正した。
 *                      以前は経度の中点が球面上に2点あることを考慮していなかった。
 *                      プロパティにHealthyを追加した。
 *                      以上の2点及び11日に追加した分の一王のテストは実施済み。
 *           2012/5/21  コメントを少しだけ見直した。
 *                      == の定義に測地系の条件も含めた。
 * ***********************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GNSS.GPS;
using GNSS.GeodeticDatum;

namespace GNSS
{
    /*2011/7/6 動作確認済み*/
    /// <summary>
    /// 緯度・経度・楕円体高を表す構造体
    /// <para>語源はドイツ語らしい。</para>
    /// <para>座標変換処理等を実装する必要もあるが、現時点ではWGS84専用となっている。</para>
    /// </summary>
    public struct BLH
    {
        /* メンバ変数 ***************************************************/
        /// <summary>緯度</summary>
        private double b;
        /// <summary>経度</summary>
        private double l;
        /// <summary>楕円体高[m]</summary>
        private double h;
        /// <summary>緯度・経度の単位</summary>
        private AngleUnit _unit;
        /// <summary>測地系</summary>
        private Datum _datum;
        /* 演算子 ***************************************************/
        /// <summary>
        /// 二項+演算子（これで足し算が簡単にできる）
        /// <para>単位はdegreeとなる</para>
        /// </summary>
        /// <param name="c1">被加算値</param>
        /// <param name="c2">加算値</param>
        /// <returns>2値を加算した結果</returns>
        public static BLH operator +(BLH c1, BLH c2)
        {
            return new BLH(c1.ToDegree().B + c2.ToDegree().B, c1.ToDegree().L + c2.ToDegree().L, c1.ToDegree().H + c2.ToDegree().H, AngleUnit.Degree);
        }
        /// <summary>
        /// 二項-演算子（これで引き算が簡単にできる）
        /// <para>単位はdegreeとなる</para>
        /// </summary>
        /// <param name="c1">被減算値</param>
        /// <param name="c2">減算値</param>
        /// <returns>2値の引き算の結果</returns>
        public static BLH operator -(BLH c1, BLH c2)
        {
            return new BLH(c1.ToDegree().B - c2.ToDegree().B, c1.ToDegree().L - c2.ToDegree().L, c1.ToDegree().H - c2.ToDegree().H, AngleUnit.Degree);
        }
        //比較演算子の==と!=をオーバーロードする 2011/12/16 ほぼコピペなので少し不安
        /// <summary>
        /// 比較演算子==
        /// <para>単位及び測地系の設定も含めて一致しなければtrueとはしません。</para>
        /// </summary>
        /// <param name="c1">比較対象その1</param>
        /// <param name="c2">比較対象その2</param>
        /// <returns>等しければture</returns>
        public static bool operator ==(BLH c1, BLH c2)
        {
            if (object.ReferenceEquals(c1, c2))                 // 両方nullか（参照元が同じか）チェックする。(c1 == c2)とすると、無限ループ
            {
                return true;
            }
            if (((object)c1 == null) || ((object)c2 == null))   // どちらかがnullかチェックする。(c1 == null)とすると、無限ループ
            {
                return false;
            }
            return (c1.B == c2.B) && (c1.L == c2.L) && (c1.H == c2.H) && (c1._unit == c2._unit) && (c1._datum == c2._datum);
        }
        /// <summary>
        /// 比較演算子!=
        /// </summary>
        /// <param name="c1">比較対象その1</param>
        /// <param name="c2">比較対象その2</param>
        /// <returns>等しくなければture</returns>
        public static bool operator !=(BLH c1, BLH c2)
        {
            return !(c1 == c2);                 // (c1 != c2)とすると、無限ループ
        }
        /// <summary>
        /// ハッシュ値を返す
        /// </summary>
        /// <returns>全メンバのXOR結果</returns>
        public override int GetHashCode()
        {
            return this.B.GetHashCode() ^ this.L.GetHashCode() ^ this.H.GetHashCode() ^ this._unit.GetHashCode();
        }
        /// <summary>
        /// objと自分自身が等価のときはtrueを返す
        /// </summary>
        /// <param name="obj">比較したいオブジェクト</param>
        /// <returns>等価であればtrue</returns>
        public override bool Equals(object obj)
        {
            //objがnullか、型が違うときは、等価でない
            if (obj == null || this.GetType() != obj.GetType())
            {
                return false;
            }
            //この型が継承できないクラスや構造体であれば、次のようにできる
            //if (!(obj is TestClass))

            // メンバで比較する
            BLH c = (BLH)obj;
            return (this.B == c.B && this.L == c.L && this.H == c.H && this.Unit == c.Unit && this.DatumKind == c.DatumKind);
        }
        /* プロパティ ***********************************************/
        /// <summary>
        /// 緯度
        /// <para>緯度は-90～90 degの範囲に丸められます。</para>
        /// <para>もしセットした値の絶対値が90°を超えている場合、極の向こう側へ移動したとみなして経度が自動的に半周します。</para>
        /// </summary>
        public double B
        {
            get { return this.b; }
            set {
                this.b = value;
                FilterAsLatitudeAndLongitude(ref this.b, ref this.l, this.Unit);
            }
        }
        /// <summary>
        /// 経度
        /// <para>経度は-180～180 degの範囲に丸められます。</para>
        /// </summary>
        public double L
        {
            get { return this.l; }
            set {
                this.l = value;
                FilterAsLongitude(ref this.l, this.Unit);
            }
        }
        /// <summary>
        /// 楕円体高
        /// <para>楕円体表面からの鉛直距離</para>
        /// </summary>
        public double H
        {
            get { return this.h; }
            set {
                double _h = value;                  // 地心よりも深い値を指定されてもそれ以上いかないようにする。地球の反対側に向かっていかせてもいいけど…。
                if (_h < -6400000) _h = -6400000;   // ほんとは-赤道半径が良いんだろうなぁ。
                this.h = _h; 
            }
        }
        /// <summary>
        /// 演算単位
        /// <para>deg or radian</para>
        /// </summary>
        public AngleUnit Unit
        {
            get 
            {
                return this._unit;
            }
            set
            {
                if (value == AngleUnit.Degree && this._unit == AngleUnit.Radian)
                {
                    this.ChangeToDegree();
                }
                else if (value == AngleUnit.Radian && this._unit == AngleUnit.Degree)
                {
                    this.ChangeToRadian();
                }
            }
        }
        /// <summary>
        /// 測地系
        /// <para>測地系の変更に対しての処理は未実装</para>
        /// </summary>
        public Datum DatumKind
        {
            get
            {
                return this._datum;
            }
            set {
                this._datum = value;
            }
        }
        /// <summary>
        /// 本オブジェクトの状態を表す
        /// <para>true: 正常, 利用可能</para>
        /// </summary>
        public Boolean Healthy
        {
            get {
                if (double.IsNaN(this.b) || double.IsNaN(this.l) || double.IsNaN(this.h))
                    return false;
                else
                    return true;
            }
        }
        /* メソッド ***********************************************/
        /// <summary>
        /// 単位を度へ変換したオブジェクトを返す
        /// </summary>
        public BLH ToDegree()
        {
            BLH ans = this;

            if (this._unit == AngleUnit.Radian)
            {
                ans.B = this.B * 180.0d / Math.PI;
                ans.L = this.L * 180.0d / Math.PI;
                ans._unit = AngleUnit.Degree;
            }
            return ans;
        }
        /// <summary>
        /// 単位をラジアン単位へ変換したオブジェクトを返す
        /// </summary>
        public BLH ToRadian()
        {
            BLH ans = this;

            if (this._unit == AngleUnit.Degree)
            {
                ans.B = this.B * Math.PI / 180.0d;
                ans.L = this.L * Math.PI / 180.0d;
                ans._unit = AngleUnit.Radian;
            }
            return ans;
        }
        /// <summary>
        /// 楕円体に沿った、東西方向の単位長[m/deg]を返す
        /// <para>実際は高度の分だけ若干の誤差が発生するし、2地点の高度差は考慮しない。</para>
        /// <para>参考：理科年表，p.563，2003．</para>
        /// <para>ちなみに、http://yamadarake.web.fc2.com/trdi/2009/report000001.html　を見ると使用した公式がヒュベニの公式と言うものであることが分かる</para>
        /// </summary>
        /// <returns></returns>
        public Length GetUnitLengthForEN(Datum datum = Datum.WGS84)
        {
            Length ans = new Length(0, 0);
            BLH temp = this.ToRadian();

            ans.E = Math.PI / 180.0 * WGS84.a * Math.Cos(temp.B) / Math.Sqrt(1.0d - WGS84.e2 * Math.Pow(Math.Sin(temp.B), 2.0d));   // 高度については無視
            ans.N = Math.PI / 180.0 * WGS84.a * (1.0d - WGS84.e2) / Math.Pow(1.0d - WGS84.e2 * Math.Pow(Math.Sin(temp.B), 2.0d), 1.5d); // 地表面に限れば誤差は無視可能。誤差は、地上では最大6408/6400程度
            return ans;
        }
        /// <summary>
        /// 自身の単位を度へ変換する
        /// </summary>
        /// 
        private void ChangeToDegree()
        {
            if (this._unit == AngleUnit.Radian)
            {
                this.B = this.B * 180.0 / Math.PI;
                this.L = this.L * 180.0 / Math.PI;
                this._unit = AngleUnit.Degree;
            }
            return;
        }
        /// <summary>
        /// 自身の単位をラジアン単位へ変換する
        /// </summary>
        private void ChangeToRadian()
        {
            if (this._unit == AngleUnit.Degree)
            {
                this.B = this.B * Math.PI / 180.0;
                this.L = this.L * Math.PI / 180.0;
                this._unit = AngleUnit.Radian;
            }
            return;
        }
        /// <summary>
        /// 2地点間の中間座標を返す
        /// <para>（現時点では）メルカトル図法上の中間点を算出します。従って極点をまたがるような2点間の中点を求める用途には向いていません。</para>
        /// <para>また、2地点間の中間地点は地球上に2点取り得ますが、2点間を結ぶ距離が短くなる方を採用しています。</para>
        /// </summary>
        /// <param name="pos">第2の座標</param>
        /// <returns>
        /// 本インスタンスと第2の座標間の中間座標
        /// <para>新しく生成する座標の単位と測地系は本オブジェクトに合わせます。</para>
        /// </returns>
        public BLH GetMedian(BLH pos)
        {
            BLH me = this;                              // 自身のコピー
            BLH he = pos;                               // 引数のコピー
            he.Unit = this.Unit;                        // 単位は統一する
            he.DatumKind = this.DatumKind;              // 測地系も統一する

            BLH relative = he - me;                     // me基準の相対座標を計算
            double centerLon = me.L + relative.L / 2.0; // 2地点の中央の経度を計算（この時点ではAbs(180.0又はπ)を超えるかもしれない）

            return new BLH((me.B + he.B) / 2.0d, centerLon, (me.H + he.H) / 2.0d, this.Unit, this.DatumKind);   // 2地点の中間座標を求める
        }
        /// <summary>
        /// 引数で渡された座標までの距離を返す
        /// <para>簡易計算なので、距離が数十km以下で有効です。</para>
        /// <para>最短距離を求めているわけではないことに注意してください。</para>
        /// </summary>
        /// <param name="pos">求めたい地点の座標</param>
        /// <returns>南北・東西方向の距離[m]を構造体で返す</returns>
        public Length GetDistance(BLH pos)
        {
            Length ans;
            BLH me = this.ToDegree();                   // 自身のコピー
            BLH he = pos.ToDegree();                    // 引数のコピー。単位はdegに統一する
            BLH center = this.GetMedian(pos);

            ans = new Length((he.L - me.L) * center.GetUnitLengthForEN().E, (he.B - me.B) * center.GetUnitLengthForEN().N); // 緯度・経度差から距離を求める。
            return ans;
        }
        /// <summary>
        /// 簡易な距離演算その2
        /// <para>
        /// 精度は、近距離では（少なくとも20km程度）その1の方法と大して変わらない。
        /// 参考：http://homepage3.nifty.com/kubota01/distance.htm
        /// 距離が50kmを超えるようなら、こちらのメソッドの方が実際の距離を表すと思う。
        /// [2011/7/18　追記] 現時点ではXYZの直交座標系への変換精度が悪く、極近距離で精度が出ない。
        /// </para>
        /// </summary>
        /// <param name="pos">求めたい地点の座標</param>
        /// <returns>距離</returns>
        public double GetDistance2(BLH pos)
        {
            XYZ me = this.ToXYZ();
            XYZ he = pos.ToXYZ();
            double d_straight = Math.Sqrt(Math.Pow(me.x - he.x, 2.0) + Math.Pow(me.y - he.y, 2.0) + Math.Pow(me.z - he.z, 2.0));   // XYZ直交座標系を用いた直線距離（線は地中に潜る）
            double Nme = WGS84.a / Math.Sqrt(1.0d - WGS84.e2 * Math.Sin(this.ToRadian().B));
            double Nse = WGS84.a / Math.Sqrt(1.0d - WGS84.e2 * Math.Sin(pos.ToRadian().B));
            double N = (Nse + Nme) / 2.0d;                                                         // 平均の半径のようなもの
            double angle = Math.Asin(d_straight / 2.0d / N);                                       // 半射程角（絵を描けば分かる）
            return 2.0d * angle * N;
        }
        /// <summary>
        /// GetDistanceを利用して、直線距離を求めるメソッド
        /// </summary>
        /// <param name="pos">求めたい地点の座標</param>
        /// <returns>距離</returns>
        public double GetDistance3(BLH pos)
        {
            Length distance0 = this.GetDistance(pos);
            double distance3 = Math.Sqrt(distance0.E * distance0.E + distance0.N * distance0.N);
            return distance3;
        }
        /// <summary>
        /// XYZ座標系へ変換したオブジェクトを返す
        /// <para>参考：理科年表，p.563，2003．</para>
        /// <para>[2011/7/18　追記]計算精度が悪い気がしてならん。</para>
        /// </summary>
        /// <param name="nextDatum">変換後の測地系</param>
        /// <returns>変換したXYZ座標</returns>
        public XYZ ToXYZ(Datum nextDatum)
        {
            XYZ xyz = new XYZ();
            double N;

            N = WGS84.a / Math.Sqrt(1.0d - WGS84.e2 * Math.Pow(Math.Sin(this.ToRadian().B), 2.0d));
            xyz.x = (N + this.H) * Math.Cos(this.ToRadian().B) * Math.Cos(this.ToRadian().L);
            xyz.y = (N + this.H) * Math.Cos(this.ToRadian().B) * Math.Sin(this.ToRadian().L);
            xyz.z = (N * (1.0d - WGS84.e2) + this.H) * Math.Sin(this.ToRadian().B);
            return xyz;
        }
        /// <summary>
        /// XYZ座標系へ変換したオブジェクトを返す
        /// <para>引数を省略できます。変換後の測地系は本オブジェクトの測地系と同じになります。</para>
        /// <para>ToXYZ(Datum nextDatum)を利用して値を返します。</para>
        /// </summary>
        /// <returns>変換したXYZ座標</returns>
        public XYZ ToXYZ()
        {
            return this.ToXYZ(this.DatumKind);
        }
        /// <summary>
        /// 緯度・経度・楕円体高の文字列化
        /// Excelに合わせて、経度から先に出力する。
        /// </summary>
        /// <returns>文字列化した位置座標</returns>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder(70);
            if (this._unit == AngleUnit.Degree)
                sb.Append(this.L.ToString("0.000000")).Append(",").Append(this.B.ToString("0.000000")).Append(",").Append(this.H.ToString("0.0"));  // 有効数字が違うことに注意
            else if (this._unit == AngleUnit.Radian)
                sb.Append(this.L.ToString("0.00000000")).Append(",").Append(this.B.ToString("0.00000000")).Append(",").Append(this.H.ToString("0.0"));
            return sb.ToString();
        }
        /// <summary>
        /// 緯度と経度にフィルタをかける
        /// <para>緯度は-90～90 degの範囲とする。経度は-180～180 degの範囲とする。</para>
        /// <example>
        /// <code>
        /// double lat = 125.3, lon = -186.36;      // 極を挟んで向こう側に行って、さらに地球を半周以上している
        /// GNSS.BLH.FilterAsLatitudeAndLongitude(ref lat, ref lon, GNSS.AngleUnit.Degree);
        /// </code>
        /// </example>
        /// </summary>
        /// <param name="lat">緯度</param>
        /// <param name="lon">経度</param>
        /// <param name="unit">単位</param>
        private static void FilterAsLatitudeAndLongitude(ref double lat, ref double lon, AngleUnit unit)
        {
            double quarter = 90.0, half = 180.0, circle = 360.0;

            if (unit == AngleUnit.Radian)
            {
                quarter = quarter * Math.PI / 180.0;
                half = half * Math.PI / 180.0;
                circle = circle * Math.PI / 180.0;
            }
            lat %= circle;                          // 地球をn周していたら、1周未満に収める
            if (lat > half)                         // 地球を半周以上していれば360°引く（又は足す）ことで±180°以内に収める
                lat -= circle;
            else if (lat < -180.0)
                lat += circle;

            if (lat > quarter || lat < -quarter)    // 極を超えているかチェック
            {
                if (lat > quarter)
                    lat = half - lat;
                else if (lat < -quarter)
                    lat = -half - lat;
                lon += half;                        // 極を挟んだ反対側に行ったわけなので、経度を180°足す
            }
            FilterAsLongitude(ref lon, unit);       // 経度に関して処理
            return;
        }
        /// <summary>
        /// 経度にフィルタをかける
        /// <para>経度は-180～180 degの範囲とする。</para>
        /// <example>
        /// <code>
        /// double lon = -186.36;                   // 地球を半周以上している
        /// GNSS.BLH.FilterAsLongitude(ref lon, GNSS.AngleUnit.Degree);
        /// </code>
        /// </example>
        /// </summary>
        /// <param name="lon"></param>
        /// <param name="unit"></param>
        private static void FilterAsLongitude(ref double lon, AngleUnit unit)
        {
            double quarter = 90.0, half = 180.0, circle = 360.0;

            if (unit == AngleUnit.Radian)
            {
                quarter = quarter * Math.PI / 180.0;
                half = half * Math.PI / 180.0;
                circle = circle * Math.PI / 180.0;
            }
            lon %= circle;                          // 地球をn周していたら、1周未満に収める
            if (lon > half)                         // 地球を半周以上していれば360°引く（又は足す）ことで±180°以内に収める
                lon -= circle;
            else if (lon < -180.0)
                lon += circle;
            return;
        }
        /// <summary>
        /// BLHのコンストラクタ
        /// <para>引数は省略も可能です。</para>
        /// </summary>
        /// <param name="B">緯度: Latitude</param>
        /// <param name="L">経度: Longitude</param>
        /// <param name="H">高度: Ellipsoidal altitude</param>
        /// <param name="unit">単位(°or rad)</param>
        /// <param name="datum">測地系<para>省略するとWGS84となる。</para></param>
        public BLH(double B = 0.0d, double L = 0.0d, double H = 0.0d, AngleUnit unit = AngleUnit.Degree, Datum datum = Datum.WGS84) 
        {
            FilterAsLatitudeAndLongitude(ref B, ref L, unit);
            this.b = B; 
            this.l = L; 
            this.h = H;
            this._unit = unit;
            this._datum = datum;
        }
    }
}
