﻿/**************************************************************************
 * WGS84.cs
 * WGS84測地系パラメータ for C#
 * 
 * 開発者　：森下功啓（K.Morhista Kumamoto-University）
 * 開発履歴：2011/7/3   整備開始
 *           2011/7/6   2地点間の距離を求める、GetDistanceのデバッグ。単位長までしか計算していなかった…。
 *           2011/7/17  XYZとBLHに二項演算子を追加
 *           2011/9/19  Open(string fname)のバグを修正
 *                      GetPositions(),GetPositioningResults()において、ubloxのNMEAフォーマットに対応（現時点での要求仕様を満たすだけなので万能ではないことに注意）
 *                      PositioningInfosのToString()に高度も含まれるように変更
 *           2011/11/11 よく分かっていないが、シリアル化した方が良さそうとのことで[Serializable]を片っ端からつけてみた。今後どうなるかは・・・よく分からない。
 *           2011/11/20 BLH構造体にToString()を追加
 *           2011/12/2  XML構文エラーがある部分を訂正した
 *           2011/12/15 BLH構造体にToString()を追加
 *           2012/4/22  コメントを少しだけ付け直した。
 *                      NMEAクラスは現時点ではバイナリの処理が不可能だが、そこんとこを何とかしたい。
 *           2012/4/24  ソースコードを分割して独立させた。
 * ***********************************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GNSS.GeodeticDatum
{
    /*2011/7/6 主要パラメータチェック済み*/
    /// <summary>
    /// WGS84座標系における各種パラメータ
    /// <para>参考：西修二郎訳，GNSSのすべて，p.265，2010.2．</para>
    /// </summary>
    [Serializable]
    public static class WGS84
    {
        // 基本定数
        /// <summary>円周率π</summary>
        public const double PI = 3.141592653589;
        /// <summary>光速[m/s]</summary>
        public const double c = 2.99792458E+08;                                        // 光速[m/s]
        // 地球に関するパラメータ
        /// <summary>
        /// Equatorial Radius 赤道半径（楕円体長半径）[m]
        /// </summary>
        public const double a = 6378137.0;
        /// <summary>
        /// 扁平率の逆数
        /// </summary>
        public const double invers_oblateness = 298.257223563;
        /// <summary>
        /// 扁平率
        /// </summary>
        public const double f = 1.0d / invers_oblateness;
        /// <summary>
        /// Short Radius 短半径
        /// </summary>
        public const double b = a * (1.0d - f);
        /// <summary>
        /// Square Eccentricity 離心率^2
        /// </summary>
        public const double e2 = f * (2.0d - f);
        /// <summary>
        /// Angular Velocity of The Earth 地球の角速度[rad/sec]
        /// </summary>
        public const double wE = 7292115E-11;
        /// <summary>
        /// Gravitational Constant of The Earth 地球重力定数[m^3/s^-2]
        /// </summary>
        public const double u = 3986004.418E+8;
        /******************** プロパティ ****************************/
        /// <summary>
        /// Eccentricity 離心率e
        /// </summary>
        public static double e { get { return Math.Sqrt(e2); } }   // プロパティの形でしか提供できない。インスタンスは必要ない。
    }
}
